import 'package:enhanced_enum/enhanced_enum.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:source_gen/source_gen.dart';

class EnumExtensionGenerator {
  final ClassElement element;
  final ConstantReader annotation;
  final EnhancedEnum config;
  final buff = StringBuffer();

  String get name => element.displayName;
  Iterable<FieldElement> get values => element.fields
      .where((f) => f.type.getDisplayString(withNullability: true) == name);
  Iterable<EnhancedEnumValue?> get valueConfigs => values
      .map((v) => v.metadata.cast<ElementAnnotation?>().firstWhere(
          (m) => m?.element?.displayName == 'EnhancedEnumValue',
          orElse: () => null))
      .map((v) => v == null
          ? null
          : EnhancedEnumValue(
              name: v.computeConstantValue()?.getField('name')?.toStringValue(),
            ));
  Map<String, EnhancedEnumValue?> get valueConfigsMap =>
      Map.fromIterables(valueNames, valueConfigs);
  Iterable<String> get valueNames => values.map((v) => v.displayName);
  String get camelToUnder => '_camelCaseToUnderscore$name';
  String get underToCamel => '_underscoreToCamelCase$name';
  String get castFirstUpper => '_castFirstCharacterUpper$name';
  String get castFirstLower => '_castFirstCharacterLower$name';

  EnumExtensionGenerator(this.element, this.annotation)
      : assert(element.kind == ElementKind.ENUM),
        config = EnhancedEnum(
          namingConvention: EnhancedEnumNamingConvention.values[annotation
                  .objectValue
                  .getField('namingConvention')
                  ?.getField('index')
                  ?.toIntValue() ??
              0],
        );

  String generate() {
    generateHelperfunctions();
    generateFromString();
    buff.writeln('extension EnhancedEnumExtension$name on $name {');
    generateName();
    generateIsMethods();
    generateWhen();
    generateMaybeWhen();
    buff.writeln('}');
    return buff.toString();
  }

  void generateHelperfunctions() {
    buff.writeln('// ignore: unused_element');
    buff.writeln('String $camelToUnder(String s, {bool lower = true}) {');
    buff.writeln(
        '  final ret = s.replaceAllMapped(RegExp(r\'([a-z])([A-Z])\'), (m) => m.group(1)! + \'_\' + m.group(2)!);');
    buff.writeln('  return lower ? ret.toLowerCase() : ret.toUpperCase();');
    buff.writeln('}');
    buff.writeln('// ignore: unused_element');
    buff.writeln('String $underToCamel(String s, {bool lower = true}) {');
    buff.writeln(
        '  final ret = s.toLowerCase().replaceAllMapped(RegExp(r\'(\\w)_(\\w)\'), (m) => m.group(1)! + m.group(2)!.toUpperCase());');
    buff.writeln('  return lower ? ret : $castFirstUpper(ret);');
    buff.writeln('}');
    buff.writeln('// ignore: unused_element');
    buff.writeln(
        'String _castFirstCharacterUpper$name(String s) => s[0].toUpperCase() + s.substring(1);');
    buff.writeln('// ignore: unused_element');
    buff.writeln(
        'String _castFirstCharacterLower$name(String s) => s[0].toLowerCase() + s.substring(1);');
  }

  void generateFromString() {
    buff.writeln('// ignore: non_constant_identifier_names');
    buff.writeln('$name? ${name}FromString(String val) {');
    if (valueConfigs.any((c) => c?.name != null)) {
      // we have name overrides, handle them first
      buff.writeln('  final override = {');
      for (final entry in valueConfigsMap.entries) {
        if (entry.value?.name != null) {
          buff.writeln('    \'${entry.value!.name}\': $name.${entry.key},');
        }
      }
      buff.writeln('  }[val];');
      buff.writeln('  if (override != null) {');
      buff.writeln('    return override;');
      buff.writeln('  }');
    }
    buff.writeln(
        '  final cmpval = val.replaceAll(\'$name.\', \'\').replaceAll(\'_\', \'\').toLowerCase();');
    buff.writeln('  return $name.values.cast<$name?>().firstWhere((v) =>');
    buff.writeln(
        '    v.toString().replaceAll(\'$name.\', \'\').replaceAll(\'_\', \'\').toLowerCase() == cmpval');
    buff.writeln('  , orElse: () => null);');
    buff.writeln('}');
  }

  void generateName() {
    buff.writeln('@override');
    buff.writeln('// ignore: override_on_non_overriding_member');
    buff.writeln('String get name {');
    if (valueConfigs.any((c) => c?.name != null)) {
      // we have name overrides, handle them first
      buff.writeln('  final override = {');
      for (final entry in valueConfigsMap.entries) {
        if (entry.value?.name != null) {
          buff.writeln('    $name.${entry.key}: \'${entry.value!.name}\',');
        }
      }
      buff.writeln('  }[this];');
      buff.writeln('  if (override != null) {');
      buff.writeln('    return override;');
      buff.writeln('  }');
    }
    buff.writeln('  final n = toString().replaceAll(\'$name.\', \'\');');
    // if we use the enhanced enum features, i.e. `.when()` here, then we can't
    // easily  rebuild the package itself. So, we'll just fall ack to a lame switch
    // instead.
    switch (config.namingConvention) {
      case EnhancedEnumNamingConvention.original:
        buff.writeln('  return n;');
        break;
      case EnhancedEnumNamingConvention.camelCase:
        buff.writeln(
            '  return n.contains(\'_\') ? $underToCamel(n) : $castFirstLower(n);');
        break;
      case EnhancedEnumNamingConvention.upperCamelCase:
        buff.writeln(
            '  return n.contains(\'_\') ? $underToCamel(n, lower: false) : $castFirstUpper(n);');
        break;
      case EnhancedEnumNamingConvention.snakeCase:
        buff.writeln('  return $camelToUnder(n);');
        break;
      case EnhancedEnumNamingConvention.screamingSnakeCase:
        buff.writeln('  return $camelToUnder(n, lower: false);');
        break;
    }
    buff.writeln('}');
  }

  void generateIsMethods() {
    for (final n in valueNames) {
      buff.writeln(
          'bool get is${n[0].toUpperCase()}${n.substring(1)} => this == $name.$n;');
    }
  }

  void generateWhen() {
    buff.writeln('T when<T>({');
    for (final n in valueNames) {
      buff.writeln('  required T Function() $n,');
    }
    buff.writeln('}) => {');
    for (final n in valueNames) {
      buff.writeln('    $name.$n: $n,');
    }
    buff.writeln('  }[this]!();');
  }

  void generateMaybeWhen() {
    buff.writeln('T maybeWhen<T>({');
    for (final n in valueNames) {
      buff.writeln('  T? Function()? $n,');
    }
    buff.writeln('  required T Function() orElse,');
    buff.writeln('}) => {');
    for (final n in valueNames) {
      buff.writeln('    $name.$n: $n,');
    }
    buff.writeln('  }[this]?.call() ?? orElse();');
  }
}
